//
//  genetic_algorithm.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#ifndef genetic_algorithm
#define genetic_algorithm

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "file_helper.h"
#include "cellular_automata.h"
#include "phenotype.h"

#include "fitness_function.h"
#include "development.h"
#include "visualization.h"

#include "timeing.h"
#include "stats.h"
#include "population.h"
#include "chance.h"
#include "mutation.h"
#include "growth.h"
#include "elitism.h"

static double fitness_demand = FITNESS_DEMAND;
static double best_fitness_all_time = 0.0;
static double current_best_fitness = 0.0;
static int current_best_fitness_index = -1;

static weighted_element * all_best = NULL;
static int all_best_size = 0;

static int * max_gsize = NULL;
static int * min_gsize = NULL;
static int gsize_length = 0;


static int wanted_attractor_value = 0;
static int relative_close_value = 0;

//Externals set
int weight_max = 0;
population * pop = NULL;

double avr_attractor = 0;
double avr_diff = 0;

//External gets
extern setting_flags flags;
extern int finished_step;
extern int max_step;
extern int nr_of_elites;

extern int add_gene_counter;


void free_genetic_algorithm(){
    for (int i = 0; i < pop->size; i++){
        
        free_ptype_mem_buffers(pop->cur[i]);
        free_automata(pop->cur[i]);
    }
    free(pop->cur);
    free(pop->new_pop);
    free(pop->fitness);
    free(all_best);
    free(max_gsize);
    free(min_gsize);
    free(pop);
}

void init_genetic_algorithm_with_add_gene(double fit_dem,
                                          population *p,
                                          int wanted_atractor,
                                          int relative_close,
                                          int add_tresh,
                                          int add_limit ){
    setup_adding_functionality(add_tresh, add_limit);

    pop = p;
    relative_close_value = relative_close;
    wanted_attractor_value = wanted_atractor;
    fitness_demand = fit_dem != 0.0 ? fit_dem : FITNESS_DEMAND;
}

/**
 * FITNESS
 */

void get_all_attractors(){
    for(int i = 0; i < pop->size; i++){
        develop_an_attractor(pop->cur[i], &(pop->fitness[i]));
    }
}

void get_all_fitness(){
    
    int biggest_difference = 0;
    for (int i = 0; i < pop->size; i++){
        int diff = fabs(wanted_attractor_value - pop->fitness[i].atr.attractor_length);
        if (biggest_difference < diff){
                biggest_difference = diff;
        }
    }
    
    double attractor_divider = 0.0;
    double rule_divider = 0.0;
    double size_divider = 0.0;
    
    for (int i  = 0; i < pop->size; i++){
        calculate_attractor_and_rule_fitness(pop->cur[i],
                                             &(pop->fitness[i]),
                                             biggest_difference,
                                             wanted_attractor_value,
                                             relative_close_value);
        
        attractor_divider += pop->fitness[i].attractor_fitness;        
        rule_divider += pop->fitness[i].compound_fitness;
        size_divider += pop->cur[i]->gtype->size;
    }
    
    for (int i  = 0; i < pop->size; i++){
        calculate_compound_fitness(&(pop->fitness[i]), attractor_divider, rule_divider, (pop->cur[i]->gtype->size / size_divider));
    }

}

void calculate_average_attractor_and_diff(){
    avr_attractor = 0;
    avr_diff = 0;
    
    for (int i = 0; i < pop->size; i++){
        avr_attractor += pop->fitness[i].atr.attractor_length;
        avr_diff += fabs(wanted_attractor_value - pop->fitness[i].atr.attractor_length);
    }
    avr_attractor /= pop->size;
    avr_diff /= pop->size;
}

void find_fittest_individual(){
    current_best_fitness = -1;
    for (int i = 0; i < pop->size; i++){
        if (current_best_fitness < pop->fitness[i].attractor_fitness){
            current_best_fitness = pop->fitness[i].attractor_fitness;
            current_best_fitness_index = i;
        }
    }
}

void weight_fitness_table(){
    int from = 0;
    for (int i = 0; i < pop->size; i++){
        pop->fitness[i].range.from = from;
        pop->fitness[i].range.to = pop->fitness[i].compound_fitness / FITNESS_GRAIN + from;
        pop->fitness[i].range.length = pop->fitness[i].range.to - from + 1;
        from = pop->fitness[i].range.to + 1;
    }
    weight_max = from;
}

int fitness_accomplished(){
    if (fitness_demand <= current_best_fitness ){
        return 1;
    }
    return 0;
}

void update_best_fitness(int generation){
    if (best_fitness_all_time < current_best_fitness){
        best_fitness_all_time = current_best_fitness;
        add_gene_counter = 0;
    }
    
    all_best[generation].atr = pop->fitness[current_best_fitness_index].atr;
    all_best[generation].attractor_fitness = pop->fitness[current_best_fitness_index].attractor_fitness;
    
    add_gene_counter ++;
}


/**
 * GENERATION SHIFT
 */
void free_old_pop_and_make_new_current(){
    for (int i = 0; i < pop->size; i++){
        free_automata(pop->cur[i]);
        pop->cur[i] = pop->new_pop[i];
    }
}

/**
 * EVOLUTION LOOP
 */
int algorithm_loop(int max_loop){
    
    //initial settings
    all_best = malloc(sizeof(weighted_element)*max_loop);
    all_best_size = 0;
    
    min_gsize = malloc(sizeof(int) * max_loop);
    max_gsize = malloc(sizeof(int) * max_loop);
    gsize_length = 0;
    
    current_best_fitness = 0;
    current_best_fitness_index = -1;
    
    best_fitness_all_time = 0.0;
    
    flags.is_successfull = 0;
    
    int generation_counter = 0;


    while (generation_counter < max_loop){

        get_all_attractors();       //Get attractors for whole population
        get_all_fitness();          //Caluclate attractor, and comul fitness for whole population
        
        start_timeing();
        weight_fitness_table();     //Weight all individuals based on comul fitness
        
        find_fittest_individual();               //Get the fittest individual
        update_best_fitness(generation_counter); //Update all best and alltime best individual

        save_timeing_to(GA);
        //Some simple output
      //  printf("\n\n+--------------+\ncur fitness %f\nbst fitness %f\n", current_best_fitness, best_fitness_all_time);
        
        /**
         * check if finished or on last round and should
         * avoid mutation and filling new population
         */
        if ( fitness_accomplished() ){
            flags.is_successfull = 1;
            generation_counter++;
            break;
        }else if (generation_counter >= (max_loop-1)){
            generation_counter++;
            break;
        }
        start_timeing();
        perform_elitism(nr_of_elites);              //Elitism before all the other, so it does not mess with the elite
        
        add_chromosome_to_pop();                    //add chromosomes if enabled
        
        fill_new_population(nr_of_elites);          //fill new pop, with crossover
        mutate_new_population(nr_of_elites, pop);   //mutation
        
        int min = 244;
        int max = 0;
        for (int i = 0; i < pop->size; i++){
            int s = pop->new_pop[i]->gtype->size;
            min = (int)MIN(min, s);
            max = (int)MAX(max, s);
        }
        max_gsize[generation_counter] = max;
        min_gsize[generation_counter] = min;
       // printf("Biggest gsize %i, smallet gsize %i\n", max, min);
        
        
        
        free_old_pop_and_make_new_current();      //free cur buffers, and make nxt cur buffers
        generation_counter ++;
        save_timeing_to(GA);
        
        if (generation_counter % 1000 == 0){
            printf("still running at counter %d\n", generation_counter);
            print_current_spent();
        }
    }
    print_time_stats();
    all_best_size = generation_counter;
    gsize_length = generation_counter;
    
    calculate_average_attractor_and_diff();

    finished_step = generation_counter;

    
    return current_best_fitness_index;
}

/**
 * Save stuff
 */
void save_stats_for_population(population * p, int best_index){
    
    open_attractor_dev_file();
    //save attractor_development
    write_attractor_development(all_best, all_best_size);
    close_attractor_dev_file();

    
    open_save_file(0);
    //save Stats
    write_statistics(&(all_best[all_best_size-1].atr) );
    //more stats
    write_more_stats();
    
    //save Genotype
    write_genotype(p->cur[best_index]->gtype, p->cur[best_index]->val.states);
    //save growth
    write_growth(min_gsize, max_gsize, gsize_length);
    close_save_file(0);
    
    //save Development
    /*
    attractor * a = run_development(p->cur[best_index], 100000, 0);
    free(a);
    */
}

#endif
