//
//  genotype.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 3/26/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "genotype.h"
#include "development.h"
#include "file_helper.h"
#include "mersenne_twister.h"
/**
 *      Gene construction
 *      11  11  11  11  11  11  11  11
 *      DC  DC  N   S   W   C   E   NEXT
 *
 */

void free_genotype(list * gtype){
    free(gtype->array);
    free(gtype->used);
    free(gtype);
}

list * init_empty_genotype(int size){
    list * l = (list*)malloc(sizeof(list));
    l->array = calloc(size, sizeof(uint32_t));
    l->used = calloc(size, sizeof(char));
    l->mods = 0;
    l->size = size;
    return l;
}

uint32_t check_skiping(automata * ca, int bits_pr_state, uint32_t current_gene){

    int mask = pow(2, bits_pr_state) - 1;
    int current_examine = current_gene & mask;
    
    if (ca->val.states  <= (current_examine + 1) ){
        current_gene |= mask;
        uint32_t ret = check_skiping(ca, bits_pr_state, current_gene >> bits_pr_state);
        
        current_gene |= ret << bits_pr_state;
    }
    
    return current_gene;
    
}

list * alloc_full_genotype(automata * ca){
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    
    int possible_perm = pow(2, bits_pr_state);
    int needs_skipping = possible_perm != ca->val.states;
    
    uint32_t current_gene = 0;
    uint32_t used_neighbourhoods = pow(ca->val.states, ca->nh_size);
    list * l = init_empty_genotype(used_neighbourhoods);
    
    for (int i = 0; i < used_neighbourhoods; i++){
        l->array[i] = current_gene << bits_pr_state;
        if (needs_skipping)
            current_gene = check_skiping(ca, bits_pr_state, current_gene);
        
        current_gene ++;
    }
    
    return l;
}

list * init_full_genotype_uniform_next(automata * ca, int next){

    list * l = alloc_full_genotype(ca);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    next &= mask;
    
    uint32_t length = pow(ca->val.states, ca->nh_size);

    for (int i = 0; i < length; i++){
        l->array[i] = l->array[i] | next;
    }

    return l;
}

list * init_full_genotype_random_next (automata * ca){

    list * l = alloc_full_genotype(ca);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    uint32_t length = pow(ca->val.states, ca->nh_size);
    
    //To make sure that all 0 dont become 1.
    int i = 0;
    l->array[i] = l->array[i] | (0 & mask);
    i ++;
    for (; i < length; i++){
        char random = genrand_int32() %  (int)ca->val.states;
        l->array[i] = l->array[i] | (random & mask);
    }
    
    return l;
}

list * init_full_genotype(automata * ca, int next){
    if (next < 0){
        return init_full_genotype_random_next(ca);
    } else {
        return init_full_genotype_uniform_next(ca, next);
    }
}


uint32_t create_chromosome_with_next(int state_n, int next){
    int bits_n = (int)ceil((double)state_n / 2.0);
    int mask = pow(2, bits_n) -1;
    
    uint32_t gene = 0;
    for (int i = 1; i < 6; i++){
        int shift = bits_n * i;
        gene |= (genrand_int32() & mask) % state_n << shift;
    }
    gene |= next & mask;
    // to make sure that all 0 dont become 1.
    if (!(gene & ~mask)){
        gene = 0;
    }
    return gene;
}

uint32_t create_random_gene(int state_n){
    int bits_n = (int)ceil((double)state_n / 2.0);
    int mask = pow(2, bits_n) -1;
    
    uint32_t gene = 0;
    for (int i = 0; i < 6; i++){
        int shift = bits_n * i;
        gene |= (genrand_int32() & mask) % state_n << shift;
    }
    // to make sure that all 0 dont become 1.
    if (!(gene & ~mask)){
        gene = 0;
    }
    return gene;
}

list * init_random_genotype (int number_of_genes, int number_of_states){
    list * l = init_empty_genotype(number_of_genes);
    for (int i = 0; i < number_of_genes; i++){
        l->array[i] = create_random_gene(number_of_states);
    }
    return l;
}

list * init_restircted_genotype_empty_genes(int genes_n){
    return init_empty_genotype(genes_n);
}

list * init_restricted_genotype (int number_of_genes, int number_of_states, int next_value){

    if ( next_value < 0){
        return init_random_genotype(number_of_genes, number_of_states);
    }else{
        list * l = init_empty_genotype(number_of_genes);
        for (int i = 0; i < number_of_genes; i++){
            l->array[i] = create_chromosome_with_next(number_of_states, next_value);
        }
        return l;
    }
}

/**
 * UTILITY
 **/
void copy_gtype(list * to, list *from){
    memcpy(to->array, from->array, (from->size * sizeof(uint32_t)));
}

list * grow_genotype(list * l){
    int new_size = (l->size + 1);
    l->array = realloc(l->array, sizeof(uint32_t) * new_size);
    l->used = realloc(l->used, sizeof(char) * new_size);
    l->size = new_size;
    return l;
}

list * grow_gene_copy_random(list * l){

    int index = genrand_int32() % l->size;
    uint32_t copied_gene = l->array[index];
    
    l = grow_genotype(l);
    l->array[l->size-1] = copied_gene;

    return l;
}





/**
 * CALCULATION
 **/
char get_next_state_apply_rules_to_gene(list * gtype, uint32_t gene, uint32_t mask, const int states){
    int found = 0;
    uint8_t sum = 0;
    
    for (int i = 0; i < gtype->size; i++){
        if (gtype->array[i] != 0 && (gtype->array[i] & ~mask) == (gene & ~mask) ){
            sum += (gtype->array[i] & mask);
            if (gtype->array[i] & ~mask){
                gtype->used[i] = 1;
            }
            found ++;
        }
    }
    
    if (found){
        gtype->mods = found - 1;
        return (char)sum % states;
    }
    
    return (char)(gene & mask);
}

/**
 * IO
 **/
void save_genotype_to_disk(list * gtype, int states){
    open_save_file(0);
    write_genotype(gtype, states);
    close_save_file(0);
}

/**
 * TESTING
 **/
list * wolfram_gtype(automata * ca, int8_t number){
    list * l = alloc_full_genotype(ca);
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        int8_t next = number & mask;
        l->array[i] = l->array[i] | next;
        number = number >> bits_pr_state;
    }
    return l;
}
list * growing_diagonal_2d(automata * ca){
    list * l = init_full_genotype(ca, NEXT_VALUE);

    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    int south = 5;
    int west = 3;
    
    uint32_t trigger = 1 << ((bits_pr_state) * south);
    trigger |= 1 << ((bits_pr_state) * west);
    uint32_t trigger1 = 1 << ((bits_pr_state) * west);
    uint32_t trigger2 = 1 << ((bits_pr_state) * south);
                    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        if ((l->array[i] & ~mask) == trigger){
            l->array[i] = l->array[i] | 1;
        }
        if ((l->array[i] & ~mask) == trigger1){
            l->array[i] = l->array[i] | 1;
        }
        if ((l->array[i] & ~mask) == trigger2){
            l->array[i] = l->array[i] | 1;
        }
    }
    
    return l;
}
list * growing_north_2d(automata * ca){
    list * l = init_full_genotype(ca, NEXT_VALUE);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    int south = 4;
    uint32_t trigger = 1 << ((bits_pr_state) * south);
    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        if ((l->array[i] & ~mask) == trigger){
            l->array[i] = l->array[i] | 1;
        }
    }
    
    return l;
}
list * growing_south_2d(automata * ca){
    list * l = init_full_genotype(ca, NEXT_VALUE);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    int north = 5;
    uint32_t trigger = 1 << ((bits_pr_state) * north);
    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        if ((l->array[i] & ~mask) == trigger){
            l->array[i] = l->array[i] | 1;
        }
    }
    
    return l;
}
list * growing_east_2d(automata * ca){
    list * l = init_full_genotype(ca, NEXT_VALUE);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    int west = 3;
    uint32_t trigger = 1 << ((bits_pr_state) * west);
    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        if ((l->array[i] & ~mask) == trigger){
            l->array[i] = l->array[i] | 1;
        }
    }
    
    return l;
}
list * growing_west_2d(automata * ca){
    list * l = init_full_genotype(ca, NEXT_VALUE);
    
    int bits_pr_state = (int)ceil((double)ca->val.states / 2.0);
    uint32_t mask = pow(2, bits_pr_state) - 1;
    
    int east = 1;
    uint32_t trigger = 1 << ((bits_pr_state) * east);
    
    for (int i = 0; i < pow(ca->val.states, ca->nh_size); i++){
        if ((l->array[i] & ~mask) == trigger){
            l->array[i] = l->array[i] | 1;
        }
    }
    
    return l;
}

