//
//  population.c
//  Experiment
//
//  Created by Andreas Giskeødegård on 4/9/13.
//  Copyright (c) 2013 Andreas Giskeødegård. All rights reserved.
//

#include <stdio.h>
#include <string.h>
#include <math.h>

#include "population.h"

#include "cellular_automata.h"
#include "genotype.h"
#include "phenotype.h"
#include "attractor_checking.h"
#include "chance.h"
#include "mersenne_twister.h"


//Extern get
extern population * pop;
extern int weight_max;

population * init_restricted_population(int pop_size, ca_values values, int buffersize, int restriction){
    population * p = malloc(sizeof(population));
    
    p->size = pop_size != 0 ? pop_size : POPULATION_SIZE;
    p->cur = calloc(p->size, sizeof(automata*));
    p->new_pop = calloc(p->size, sizeof(automata*));
    p->fitness = calloc(p->size, sizeof(weighted_element));
    
    for (int i = 0; i < p->size; i++){
        p->cur[i] = init_restricted_automata_with(values, buffersize, restriction);
    }
    return p;
}

population * init_growing_population(int pop_size, ca_values values, int buffersize, int start_size){
    population * p = malloc(sizeof(population));
    
    p->size = pop_size != 0 ? pop_size : POPULATION_SIZE;
    p->cur = calloc(p->size, sizeof(automata*));
    p->new_pop = calloc(p->size, sizeof(automata*));
    p->fitness = calloc(p->size, sizeof(weighted_element));
    
    for (int i = 0; i < p->size; i++){
        p->cur[i] = init_growing_automata_with(values, buffersize, start_size);

    }
    return p;
}


population * init_population(int population_size, ca_values values, int buffersize){
    population * p = malloc(sizeof(population));
    p->size = population_size != 0 ? population_size : POPULATION_SIZE;

    p->cur = calloc(p->size, sizeof(automata*));
    p->new_pop = calloc(p->size, sizeof(automata*));
    p->fitness = calloc(p->size, sizeof(weighted_element));
    
    for (int i = 0; i < p->size; i++){
        p->cur[i] = init_random_automata_with(values, buffersize);
    }
    return p;
}


/**
 * BREEDING
 */


int value_in_range(int selected_value, range r){
    return r.from <= selected_value && selected_value <= r.to;
}

int reduce_selected_value(int selected_value, range r){

    while (value_in_range(selected_value, r)) {
        selected_value = r.to + 1 + (selected_value - r.from); // pluss 1 since inrange is from<=val<=to
        selected_value = selected_value % weight_max;
    }
    return selected_value;
}

automata * get_a_parent(int * selected_index, int skip_index){
    
    int selected_value = genrand_int32() % weight_max;
    if (0 <= skip_index){
        selected_value = reduce_selected_value(selected_value, pop->fitness[skip_index].range);
    }
    
    for (int i = 0; i < pop->size; i++){
        
        weighted_element elemt = pop->fitness[i];
        int in_range = value_in_range(selected_value, elemt.range);
        
        if (i != skip_index && in_range){
            if (selected_index != NULL){
                *selected_index = i;
            }
            return pop->cur[i];
        }
    }
    
    printf("getting a parent failed \nselected value is %i \nwhere weight max is %i\nexiting...\n", selected_value, weight_max);
    exit(1);
}

int add_automata_with_genotype(int * index, ca_values values, list * gtype){
    if (*index < pop->size){
        
        pop->new_pop[*index] = init_automata(values);
        pop->new_pop[*index]->gtype = gtype;
        (*index) ++;
        return 1;
    }else{
        free_genotype(gtype);
        return 0;
    }
}

int fill_two_new_children_from_index(int * index){
    int selected_index = -1;
    
    automata * parent_a = get_a_parent(&selected_index, -1);
    automata * parent_b = get_a_parent(NULL, selected_index);
    
    list * prnt_gt_a = parent_a->gtype;
    list * prnt_gt_b = parent_b->gtype;
    
    list * child_a = init_empty_genotype(prnt_gt_a->size);
    list * child_b = init_empty_genotype(prnt_gt_b->size);
    
    copy_gtype(child_a, prnt_gt_a);                         //clone parent a
    copy_gtype(child_b, prnt_gt_b);                         //clone parent b
    
    if (hit_on_probability(CROSSOVER_RATE)){                    //check for crossover
        int min = MIN(prnt_gt_a->size, prnt_gt_b->size);    //get smallest genome size
        int length = (genrand_int32() % min) + 1;                    //find length of chromosomes to cross
        
        int pos = min - length;                             //find start poss for xover
        length = (sizeof(uint32_t) * length);
        
        memcpy(&(child_a->array[pos]), &(prnt_gt_b->array[pos]), length); //xover parent b to child a
        memcpy(&(child_b->array[pos]), &(prnt_gt_a->array[pos]), length); //xover parent a to child b
        //safety not needed because of the way we split the gtype horizontally.
    }
    
    if (!add_automata_with_genotype(index, parent_a->val, child_a)){    //add child a to nxtgen if space
        free_genotype(child_b);
        return 0;
    }
    return add_automata_with_genotype(index, parent_a->val, child_b);   //add child b to nxtgen if space
}

void transfer_phenotype_buffers(population * pop){
    for (int i = 0; i < pop->size; i++){ // reuse the phenotype buffers and pointers
        automata * ca_a = pop->new_pop[i];
        automata * ca_b = pop->cur[i];
        
        ca_a->ptype = init_empty_ptype(ca_b);
        ca_a->ptype->crnt = ca_b->ptype->crnt;
        ca_a->ptype->nxt = ca_b->ptype->nxt;
        ca_a->ptype->indic = ca_b->ptype->indic;
        ca_a->ptype->nr_buffs = ca_b->ptype->nr_buffs;
        ca_a->ptype->buffers = ca_b->ptype->buffers;
    }
}

void fill_new_population(int number_of_elites){
    int index = number_of_elites;
    while (fill_two_new_children_from_index(&index)){};
    transfer_phenotype_buffers(pop);
}







